@@ -539,14 +539,14 @@ test_that("StructArray methods", {
   expect_equal(a$x, arrow_array(df$x))
   expect_equal(a[["x"]], arrow_array(df$x))
   expect_equal(a[[1]], arrow_array(df$x))
-  expect_identical(names(a), c("x", "y", "z"))
+  expect_named(a, c("x", "y", "z"))
   expect_identical(dim(a), c(10L, 3L))
 })
 
 test_that("StructArray creation", {
   # from data.frame
   a <- StructArray$create(example_data)
-  expect_identical(names(a), c("int", "dbl", "dbl2", "lgl", "false", "chr", "fct"))
+  expect_named(a, c("int", "dbl", "dbl2", "lgl", "false", "chr", "fct"))
   expect_identical(dim(a), c(10L, 7L))
   expect_r6_class(a, "StructArray")
 
@@ -1261,7 +1261,7 @@ test_that("concat_arrays works", {
 
   concat_int <- concat_arrays(arrow_array(1:3), arrow_array(4:5))
   expect_true(concat_int$type == int32())
-  expect_true(all(concat_int == arrow_array(1:5)))
+  expect_equal(concat_int,  arrow_array(1:5))
 
   concat_int64 <- concat_arrays(
     arrow_array(1:3),
@@ -1269,7 +1269,7 @@ test_that("concat_arrays works", {
     type = int64()
   )
   expect_true(concat_int64$type == int64())
-  expect_true(all(concat_int == arrow_array(1:5)))
+  expect_equal(concat_int, arrow_array(1:5))
 
   expect_error(
     concat_arrays(
@@ -1283,7 +1283,7 @@ test_that("concat_arrays works", {
 test_that("concat_arrays() coerces its input to Array", {
   concat_ints <- concat_arrays(1L, 2L)
   expect_true(concat_ints$type == int32())
-  expect_true(all(concat_ints == arrow_array(c(1L, 2L))))
+  expect_equal(concat_ints, arrow_array(c(1L, 2L)))
 
   expect_error(
     concat_arrays(1L, "not a number", type = int32()),
